from pwn import *
from src.pwno import *

context.log_level = 'debug'
config.ATTACHMENT = './pwn'

def testFrame():
    # test if get_instance() returns the same instance while in different stack frame
    print(">>> TESTING FRAME")
    testFrameSH: process = gen_sh()
    def send_here():
        _, my_sh = get_instance()
        assert my_sh == testFrameSH
        sl(b'0')
    
    send_here()
    testFrameSH.close()
    print(">>> FRAME TEST PASSED")

def testAbbr():
    # test if abbr works
    print(">>> TESTING ABBR")
    sh: process = gen_sh()
    my_sh_name, my_sh = get_instance()
    sl(b'0')
    assert my_sh == sh, f'sh is {sh} with PID {sh.proc.pid} but get_instance() returns {(my_sh_name, my_sh)} with PID {my_sh.proc.pid})'
    sh.close()
    print(">>> ABBR TEST PASSED")

def testAbbrAfterClose():
    print(">>> TESTING ABBR AFTER CLOSE")
    sh: process = gen_sh()
    old = sh
    sl(b'0')
    sh.close()
    
    sh: process = gen_sh()

    new = get_instance()
    assert old != new, "get_instance() is still using the old instance"
    sl(b'0')
    sh.close()
    another_name: process = gen_sh()
    new_ = get_instance()
    assert new != new_, "get_instance() is still using the old instance without changing it's name"
    sl(b'0')
    another_name.close()
    print(">>> ABBR AFTER CLOSE TEST PASSED")

def testCustomAbbr():
    # test if custom abbr works
    print(">>> TESTING CUSTOM ABBR")
    p = gen_sh()
    sh = gen_sh()

    my_sl = abbr(p.sendline)
    p_instance = get_instance('p')
    sh_instance = get_instance()

    assert p_instance != sh_instance, "get_instance() doesn't return two instances"
    my_sl(b'0')
    my_sl(b'1')
    sl(b'1')
    p.close()
    p = gen_sh()
    my_sl(b'2')
    assert p_instance != get_instance('p'), "get_instance(name) doesn't return two different instances."
    p.close()
    sh.close()
    print(">>> CUSTOM ABBR TEST PASSED")

def testAbbrRemote():
    sh = remote('localhost', 19919)
    sl(b'echo 1 > /dev/null')
    p = remote('localhost', 19919)
    sl(b'cat /flag')
    ia()

testFrame()
testAbbr()
testAbbrAfterClose()
testCustomAbbr()

testAbbrRemote()


from typing import Literal, Any
from pwn import *
from ..context import config, get_instance

DBG_CNT = -1

def uu64(data: bytes, endianness: Literal['little', 'big'] = None, sign: bool = None, **kwargs: Any):
    """
    Unpacks 64-bit integer from padded data.

    endianness and signedness is done according to context.

    Arguments:
        data (bytes): Data to unpack from
        endianness (str): Endianness of the integer ("little"/"big")
        sign (bool): Signedness of the integer (False/True)
        kwargs: Anything that can be passed to context.local

    Returns:
        The unpacked integer.
    """
    return pwnlib.util.packing.u64(data.ljust(8, b'\x00'), endianness=endianness, sign=sign, **kwargs)


def uu32(data: bytes, endianness: Literal['little', 'big'] = None, sign: bool = None, **kwargs: Any):
    """
    Unpacks 32-bit integer from padded data.

    endianness and signedness is done according to context.

    Arguments:
        data (bytes): Data to unpack from
        endianness (str): Endianness of the integer ("little"/"big")
        sign (bool): Signedness of the integer (False/True)
        kwargs: Anything that can be passed to context.local

    Returns:
        The unpacked integer.
    """
    return pwnlib.util.packing.u32(data.ljust(4, b'\x00'), endianness=endianness, sign=sign, **kwargs)


def dbg(gdb_script: str = None,
        sh: process | remote = None,
        s: int = 4, force: bool = False):
    """
    附加 GDB，默认选择最新的 process
    参数：
        gdb_script(str)
            GDB 调试参数，用 \n 分割
        sh(process | remote)
            指定附加的 sh
        s(int)
            等待时间，在 attach 上进程后等待多少秒，默认为 4s
        force(bool)
            强制调试 gdb.debug() 的实例或是 remote 实例
    """
    if (config.REMOTE or config.GDB or config.NO_DEBUG) and not force:
        return
    if sh is None:
        _, sh = get_instance()
        if isinstance(sh, remote) and not force:
            return
        
    global DBG_CNT
    DBG_CNT += 1
    if config.DBG and DBG_CNT not in config.DBG:
        return
    gdb.attach(sh, gdb_script)
    if s != 0 and not config.GDB:
        pause(s)



import subprocess
from pathlib import Path
from pwn import *
from elftools.common.exceptions import ELFError
from argparse import ArgumentParser
from pydantic import BaseModel, root_validator, ValidationError, validator

# ------- Default Settings -------
class Config(BaseModel, extra='ignore'):
    ATTACHMENT: str = None
    RUNARGS: str = ''
    LIBC: str = '/lib/x86_64-linux-gnu/libc.so.6'
    HOST: str = ''
    PORT: int = 0

    NO_DEBUG: bool = False
    REMOTE: bool = False
    GDB: bool = False  # gdb.debug(elf.path, gdbscript=gdbscript)
    GDB_SCRIPT: str = ''
    DBG: list[int] = None

    @root_validator(pre=True)
    def ignore_None(cls, values):
        fin = {k: v for k, v in values.items() if v is not None}
        keys = list(fin.keys())
        if ('HOST' in keys and 'PORT' not in keys) \
            or \
            ('PORT' in keys and 'HOST' not in keys):
            raise ValidationError(f"Host and Port should be both set.")


        if 'REMOTE' in fin.keys():
            rmt = values['REMOTE'].split(':')
            fin['HOST'] = rmt[0]
            fin['PORT'] = int(rmt[1])
            fin['REMOTE'] = True
        return fin

    @validator('ATTACHMENT', pre=True, always=True)
    def _attachment(cls, value) -> str:
        if value is None:
            def is_elf(path: Path):
                try:
                    result = subprocess.run(['file', str(path)], capture_output=True, text=True)
                    return "ELF" in result.stdout and "executable" in result.stdout
                except:
                    return False
            for file in Path.cwd().iterdir():
                if file.is_file() and is_elf(file):
                    info("No Attachment set, using \"%s\"...", file.absolute())
                    return file.name
            return '/bin/sh'  # fallback
        return value
    
    @validator('DBG', pre=True, always=True)
    def _dbg(cls, value) -> list[int]:
        if value is None:
            return []
        return [int(i) for i in value.split(',')]

parser = ArgumentParser(description="Pwnable Commandline")
parser.add_argument('ATTACHMENT', nargs='?')
parser.add_argument('--libc', '-l', nargs='?', dest='LIBC')
parser.add_argument('--debug', '-d', action='store', dest='DBG', help='Which dbg() to be executed, default is `all`, use comma to split. e.g. `-d 0,1,3`')
parser.add_argument('--no-debug', '-D', action='store_true', dest='NO_DEBUG', help='Disable debug mode')
parser.add_argument('--remote', '-r', action='store', dest='REMOTE', help='Remote host:port')
parser.add_argument('--host', '-H', action='store', dest='HOST', help='Remote host, if remote is set, this option will be ignored')
parser.add_argument('--port', '-p', action='store', dest='PORT', help='Remote port, if remote is set, this option will be ignored')
parser.add_argument('--gdb', '-g', action='store_true', dest='GDB', help='Run binary using gdb.debug')
parser.add_argument('--gdb-script', '-G', action='store', dest='GDB_SCRIPT', help='GDB script to run')
parser.add_argument('--args', '-a', action='store', dest='RUNARGS', help='Arguments to run binary')
args = parser.parse_args()

config = Config(**vars(args))
print(config)
try:
    elf = ELF(config.ATTACHMENT)
except ELFError:
    elf = None
    log.warning(f"{config.ATTACHMENT} is not a valid ELF file!, `elf` is not set")
libc = ELF(config.LIBC)
context.log_level = 'debug' if not config.NO_DEBUG else 'info'


import os
from pwn import *
from .context import config

def gen_sh(*a, f_or_h: str | Path = None, port: int = None, **kw) -> process | remote:
    """
    生成一个 process 或 remote 实例，如果 config.REMOTE 为 True 则生成 remote 实例，否则生成 process 实例。
    如果传入的参数只有一个，则会将其作为 config.ATTACHMENT，如果检测到
    如果传入的参数有两个，则会将第一个参数作为 config.HOST，第二个参数作为 config.PORT。
    
    """
    if len(a) == 1 or (f_or_h is not None and port is None):
        s = f_or_h or a[0]
        if ":" in s and len(r := s.split(':')) == 2 and r[1].isdigit():
            config.HOST = r[0]
            config.PORT = int([1])
            config.REMOTE = True
        else:
            config.ATTACHMENT = s if isinstance(s, str) else str(s)
        
        if len(a) >= 1:
            a = a[1:]
    
    elif len(a) == 2 or (f_or_h is not None and port is not None):
        s = f_or_h or a[0]
        p = port or a[1]
        config.HOST = s
        config.PORT = p
        config.REMOTE = True

        if len(a) >= 2:
            a = a[2:]


    if config.REMOTE:
        return remote(config.HOST, config.PORT)
    if not config.ATTACHMENT.startswith('.') or not config.ATTACHMENT.startswith('/'):
            config.ATTACHMENT = './' + config.ATTACHMENT
    if config.GDB:
        return gdb.debug([config.ATTACHMENT, *config.RUNARGS.split(' ')], gdbscript=config.GDB_SCRIPT, *a, **kw)
    return process([config.ATTACHMENT, *config.RUNARGS.split(' ')], *a, **kw)


def initialization():
    # if we're using wsl, set context terminal to cmd.exe
    if "WSL_DISTRO_NAME" in os.environ:
        args = ['cmd.exe', '/c', 'start']
        if 'WT_SESSION' in os.environ:
                args.extend(['wt.exe', '-w', '0'])

                if distro_name := os.getenv('WSL_DISTRO_NAME'):
                    args.extend(['wsl.exe', '-d', distro_name, 'bash', '-c'])
                else:
                    args.extend(['bash.exe', '-c'])

        context.terminal = args

initialization()
